MODULE Modules;

IMPORT SYSTEM, Trace, Kernel, Files;

CONST
  done             = 0;
  fileNotFound     = 1;
  invalidObjFile   = 2;
  corruptedObjFile = 4;
  cmdNotFound      = 5;
  moduleNotFound   = 6;
  notEnoughSpace   = 7;
  refCntNotZero    = 8;
  cyclicImport     = 9;
  incompImport     = 16;
  ExtTabWordSize   = 16;
  Tag0WordOffset   = -2;
  Mth0WordOffset   = Tag0WordOffset - ExtTabWordSize;
  ActiveTag        = 0AFX;

  EUEnd            = 0;
  EURecord         = 1;
  EUobjScope       = 0;
  EUrecScope       = 1;
  EUerrScope       = -1;
  EUProcFlag       = 80000000H;

TYPE
  Command*         = Kernel.Proc;
  Module*          = Kernel.Module;
  ModuleName*      = Kernel.Name;
  ADDRESS          = LONGINT;
  ExportPtr        = Kernel.ExportPtr;
  ExportDesc       = Kernel.ExportDesc;
  Cmd              = Kernel.Cmd;

VAR
  res*: INTEGER;
  resMsg*: ARRAY 256 OF CHAR;
  extension*: ARRAY 8 OF CHAR;
  resMsgLen: LONGINT;

  imported, importing: ModuleName;
  moduleCS: Kernel.CriticalSection;

TYPE
  Header = RECORD
    nofEntries   : INTEGER;
    nofCmds      : INTEGER;
    nofPtrs      : INTEGER;
    nofTds       : INTEGER;
    nofImps      : INTEGER;
    nofLinks     : INTEGER;
    nofDataLinks : INTEGER;
    codeSize, dataSize, refSize, conSize: LONGINT;
  END;

  LinkTab = POINTER TO ARRAY OF RECORD
    mod, entry: CHAR;
    link: LONGINT
  END;

  DataLinkTab = POINTER TO ARRAY OF RECORD
    mod: CHAR;
    entry: INTEGER;
    nofFixups: INTEGER;
    offset: POINTER TO ARRAY OF LONGINT
  END;

PROCEDURE LSW (x: LONGINT): LONGINT;
BEGIN
  RETURN SYSTEM.VAL(LONGINT, SYSTEM.VAL(SET, x) * {0..15})
END LSW;

PROCEDURE MSW (x: LONGINT): LONGINT;
BEGIN
  RETURN SYSTEM.LSH(x, -16)
END MSW;

PROCEDURE ReadUnsigned (VAR File: Files.File; VAR u: LONGINT);
VAR
  low, high: CHAR;
BEGIN
  File.Read(low);
  File.Read(high);
  u := 256*LONG(ORD(high)) + ORD(low);
END ReadUnsigned;

PROCEDURE Ch(ch: CHAR);
BEGIN
  IF resMsgLen # LEN(resMsg)-1 THEN
    resMsg[resMsgLen] := ch;
    INC(resMsgLen);
    resMsg[resMsgLen] := 0X
  END
END Ch;

PROCEDURE Str(s: ARRAY OF CHAR);
VAR i: LONGINT;
BEGIN
  i := 0;
  WHILE s[i] # 0X DO
    Ch(s[i]);
    INC(i)
  END;
END Str;

PROCEDURE Ln;
BEGIN
  Ch(0DX)
END Ln;

PROCEDURE Clear;
BEGIN
  resMsg[0] := 0X;
  resMsgLen := 0
END Clear;

PROCEDURE ErrMsg(res: INTEGER);
BEGIN
  IF res # 0 THEN
    Clear;
    Str(importing);
    IF (res = fileNotFound) OR (res = moduleNotFound) THEN Str(" not found")
    ELSIF res = invalidObjFile THEN Str(" not an obj-file")
    ELSIF res = corruptedObjFile THEN Str(" corrupted obj file")
    ELSIF res = notEnoughSpace THEN Str(" not enough space")
    ELSIF res = refCntNotZero THEN Str(" reference count not zero")
    ELSIF res = cyclicImport THEN Str(" cyclic import")
    ELSIF res = incompImport THEN
      Str(" imports incompatible version of ");
      Str(imported);
    ELSIF res = cmdNotFound THEN
      Clear;
      Str(imported);
      Ch(".");
      Str(importing);
      Str(" command not found");
    ELSE
      Str(" unknown error code")
    END;
    Trace.String("Modules: ");
    Trace.String(resMsg);
    Trace.Ln
  END
END ErrMsg;



PROCEDURE FindMod*(name: ARRAY OF CHAR): Module;
VAR
  m: Module;
BEGIN
  m := Kernel.modules;
  WHILE (m # NIL) & (m.name # name) DO m := m.next END;
  RETURN m
END FindMod;

PROCEDURE And(x, y: LONGINT): LONGINT;
BEGIN
  RETURN SYSTEM.VAL(LONGINT, SYSTEM.VAL(SET, x) * SYSTEM.VAL(SET, y))
END And;

PROCEDURE ^Load (name: ARRAY OF CHAR; VAR m: Module);

PROCEDURE LoadModule (VAR File: Files.File; VAR m: Module);
TYPE
  TDesc = POINTER TO RECORD word: ARRAY 32000 OF LONGINT END;

VAR
  ch: CHAR;
  i, mno, j, symSize, e, k, t: LONGINT;
  imp: Module;
  body: Command;
  linkTab: LinkTab;
  dataLinks: DataLinkTab;
  head: Header;

  types: POINTER TO ARRAY OF RECORD
    initialized: BOOLEAN;
    tdesc: TDesc;
    entry, root, nofMeth, nofInhMeth, baseMod: INTEGER;
    baseEntry: LONGINT
  END;

  mods: POINTER TO ARRAY OF ModuleName;

  PROCEDURE FixupCall(code, link, fixval: LONGINT);
  VAR
    instr, nextlink, jmp: LONGINT;
  BEGIN
    REPEAT
      SYSTEM.GET(code + link, instr);
      nextlink := MSW(instr);
      SYSTEM.GET(code + link - 1, jmp);
      IF LSW(jmp) MOD 100H = 0E8H THEN
        SYSTEM.PUT(code + link, fixval - (code + link + 4))
      ELSE
        SYSTEM.PUT(code + link, fixval)
      END;
      link := nextlink
    UNTIL link = 0FFFFH
  END FixupCall;

  PROCEDURE FixupVar(code, link, fixval: LONGINT);
  VAR nofFixups: INTEGER; i, val: LONGINT;
  BEGIN
    nofFixups := dataLinks[link].nofFixups;
    i := 0;
    WHILE i < nofFixups DO
      SYSTEM.GET(code + dataLinks[link].offset[i], val);
      SYSTEM.PUT(code + dataLinks[link].offset[i], val + fixval);
      INC(i)
    END
  END FixupVar;

  PROCEDURE Fixup(m: Module; link: LinkTab);
  VAR
    i, val, offs: LONGINT;
    modNo, nofFixups: INTEGER;
    codebase: ADDRESS;
    Error: ARRAY 32 OF CHAR;
    Num: LONGINT;
  BEGIN
    codebase := SYSTEM.ADR(m.code[0]);
    modNo := ORD(dataLinks[0].mod);
    nofFixups := dataLinks[0].nofFixups;
    IF (head.nofDataLinks > 0) & (modNo = 0) THEN
      i := 0;
      WHILE i < nofFixups DO
        SYSTEM.GET(codebase + dataLinks[0].offset[i], val);
        SYSTEM.PUT(codebase + dataLinks[0].offset[i], val + m.sb);
        INC(i)
      END;
    END;

    i := 0;

    WHILE i < head.nofLinks DO
      IF ORD(link[i].mod) = 0 THEN
        CASE ORD(link[i].entry) OF
        |  255:
	        offs := link[i].link;
	        WHILE offs # 0FFFFH DO
		  SYSTEM.GET(m.sb + offs, val);
                  SYSTEM.PUT(m.sb + offs, codebase + LSW(val));
		  offs := MSW(val);
	        END;
        |  254:
	        offs := link[i].link;
	        WHILE offs # 0FFFFH DO
		  SYSTEM.GET(codebase + offs, val);
                  SYSTEM.PUT(codebase + offs, m.entries[LSW(val)]);
		  offs := MSW(val);
	        END;
        |  253:
	        offs := link[i].link;
	        WHILE offs # 0FFFFH DO
		  SYSTEM.GET(codebase + offs, val);
		  SYSTEM.PUT(codebase + offs, Kernel.runtime[0] - (codebase + offs + 4));
		  offs := MSW(val);
	        END;
        |   252:
	        offs := link[i].link;
	        WHILE offs # 0FFFFH DO
		  SYSTEM.GET(codebase + offs, val);
                  SYSTEM.PUT(codebase + offs, Kernel.runtime[1] - (codebase + offs + 4));
		  offs := MSW(val);
	        END;
        |   251:
	        offs := link[i].link;
	        WHILE offs # 0FFFFH DO
		  SYSTEM.GET(codebase + offs, val);
                  SYSTEM.PUT(codebase + offs, Kernel.runtime[2] - (codebase + offs + 4));
		  offs := MSW(val);
	        END;

        ELSE
          Error := "Unsupported kernel procedure";
          Num := ORD(link[i].entry);
          HALT(98);
        END;
      ELSE
        HALT(99);
      END;
      INC(i);
    END;

  END Fixup;

  PROCEDURE LoadExpBlock(VAR File: Files.File; M: Module);
  VAR
    struct, old: POINTER TO ARRAY OF LONGINT;
    nofStr: LONGINT;
    exp: ExportPtr;

    PROCEDURE LoadScope(VAR scope: ExportDesc; level: INTEGER; adr: LONGINT);
    VAR
      no, no2: INTEGER;
      fp, off: LONGINT;
    BEGIN
      File.ReadBytes(scope.nofExp, 2);
      no := 0;
      no2 := 0;
      IF scope.nofExp # 0 THEN
        NEW(scope.dsc, scope.nofExp);
        scope.dsc[0].adr := adr
      END;

      IF level = EUrecScope THEN
        INC(nofStr);
        IF nofStr = LEN(struct) THEN
	  old := struct;
	  NEW(struct, 2*nofStr);
	  FOR off := 0 TO nofStr-1 DO struct[off] := old[off] END
        END;
        struct[nofStr] := SYSTEM.ADR(scope)
      END;

      File.ReadNum(fp);
      WHILE fp#EUEnd DO
        IF fp=EURecord THEN
          File.ReadNum(off);
          IF off < 0 THEN
	    exp := SYSTEM.VAL(ExportPtr, struct[-off]);
	    scope.dsc[no2].nofExp := exp.nofExp;
            scope.dsc[no2].dsc := exp.dsc;
          ELSE
            LoadScope(scope.dsc[no2], EUrecScope, off)
          END;
        ELSE
          IF level = EUobjScope THEN File.ReadNum(scope.dsc[no].adr) END;
          scope.dsc[no].fp := fp;
          no2 := no;
          INC(no);
        END;
        File.ReadNum(fp)
      END
    END LoadScope;

  BEGIN
    NEW(struct, 64);
    nofStr := 0;
    LoadScope(M.export, EUobjScope, 0)
  END LoadExpBlock;

  PROCEDURE CheckUseBlock(VAR File: Files.File; M: Module);
  VAR
    mod: Module;
    name: ModuleName;

    PROCEDURE Err;
    BEGIN
      IF res = 0 THEN
        COPY(mod.name, imported);
        res := incompImport;
        ErrMsg(res)
      ELSE
        Ch(9X)
      END;
    END Err;

    PROCEDURE CheckScope(scope: ExportDesc; level: INTEGER);
    VAR
      fp, link, i, tdadr: LONGINT;
      tadr: ExportPtr;
      tmpErr: BOOLEAN;
    BEGIN
      tmpErr := (level = EUerrScope);
      i := 0;
      File.ReadNum(fp);
      WHILE fp # EUEnd DO
        IF fp = EURecord THEN
          File.ReadNum(link);
          IF tmpErr THEN
            CheckScope(scope.dsc[i], EUerrScope)
          ELSE
            IF scope.dsc[i].dsc # NIL THEN
              IF link # 0 THEN
                tadr := SYSTEM.VAL(ExportPtr, SYSTEM.ADR(scope.dsc[i].dsc[0]));
                SYSTEM.GET(mod.sb+tadr.adr, tdadr);
                SYSTEM.PUT(M.sb-link, tdadr);
              END
            END;
            CheckScope(scope.dsc[i], EUrecScope)
          END
        ELSE
          File.ReadString(name);
          IF level >= EUobjScope THEN
            tmpErr := FALSE;
            IF level = EUobjScope THEN File.ReadNum(link) END;
            i := 0;
            WHILE (i < scope.nofExp) & (scope.dsc[i].fp # fp) DO INC(i) END;
            IF i >= scope.nofExp THEN
              Err;
              tmpErr := TRUE;
              Ln;
              Ch(9X);
	      IF name = "@" THEN
                Str("RECORD ")
	      ELSE
                Str(name)
	      END;
	      Str(" incompatible");
	      DEC(i);
            ELSIF (level = EUobjScope) & (link # 0) THEN
              IF And(link, EUProcFlag) = 0 THEN
                FixupVar(SYSTEM.ADR(M.code[0]), link, mod.sb+scope.dsc[i].adr)
              ELSE
                FixupCall(SYSTEM.ADR(M.code[0]), link-EUProcFlag, scope.dsc[i].adr+SYSTEM.ADR(mod.code[0]))
              END;
            END;
          END;
        END;
        File.ReadNum(fp);
      END
    END CheckScope;

  BEGIN
    File.ReadString(name);
    WHILE (name # "") & (res = 0) DO
      Load(name, mod);
      IF res = 0 THEN CheckScope(mod.export, EUobjScope) END;
      File.ReadString(name)
    END
  END CheckUseBlock;

  PROCEDURE Expect (tag: CHAR): BOOLEAN;
  VAR
    ch: CHAR;
  BEGIN
    File.Read(ch);
    IF ch # tag THEN
      res := corruptedObjFile;
      ErrMsg(res);
      RETURN FALSE
    ELSE
      RETURN TRUE
    END
  END Expect;

  PROCEDURE FindTDescAdr(M: Module; fp: LONGINT; VAR adr: LONGINT);
  VAR
    i: INTEGER;
  BEGIN
    i := 0;
    WHILE ( i < M.export.nofExp) & (M.export.dsc[i].fp # fp) DO INC(i) END;
    IF i < M.export.nofExp THEN
      adr := M.export.dsc[i].dsc[0].adr
    ELSE
      HALT(99)
    END
  END FindTDescAdr;

  PROCEDURE ReadTypes (VAR File: Files.File);
  VAR
    i, j, k, tdsize, recordSize: LONGINT;
    tdEntry, nofMethods, nofNewMeths, mthNo, nofPointers, root, entryNo: INTEGER;
    td: TDesc;
    name: ModuleName;
    ch: CHAR;
  BEGIN
    IF LEN(m.tdescs^) # 0 THEN
      NEW(types, LEN(m.tdescs^));
      FOR i := 0 TO LEN(m.tdescs^)-1 DO
        types[i].initialized := FALSE;
        File.ReadLInt(recordSize);
        File.ReadInt(tdEntry); types[i].entry := tdEntry;
        File.ReadInt(types[i].baseMod);
        File.ReadLInt(types[i].baseEntry);
        File.ReadInt(nofMethods);
        types[i].nofMeth := nofMethods;
        File.ReadInt(types[i].nofInhMeth);
        File.ReadInt(nofNewMeths);
        File.ReadInt(nofPointers);
        File.ReadString(name);
        root := 13 + nofMethods + ExtTabWordSize + 1;
        INC(root, (-root+2) MOD 4);
        types[i].root := root;
        tdsize := (root + 1 + nofPointers + 1)*4;
        SYSTEM.NEW(td, tdsize - 24);
        DEC(SYSTEM.VAL(ADDRESS, td), 24);
        types[i].tdesc := td;
        td.word[0] := tdsize;
        td.word[1] := -4;
        td.word[2] := SYSTEM.ADR(td.word[root]);
        td.word[3] := 0;
        k := SYSTEM.ADR(td.word[4]);
        j := 0;
        REPEAT
          ch := name[j];
          SYSTEM.PUT(k, ch);
          INC(j);
          INC(k)
        UNTIL ch = 0X;
        td.word[12] := SYSTEM.ADR(m^);
        j := 0;
        WHILE j < nofNewMeths DO
	  File.ReadInt(mthNo);
          File.ReadInt(entryNo);
	  td.word[root + Mth0WordOffset - mthNo] := m.entries[entryNo];
	  INC(j)
        END;
        ASSERT(SYSTEM.ADR(td.word[0]) # 0);
        td.word[root - 1] := SYSTEM.ADR(td.word[0]);
        td.word[root] := recordSize;
        SYSTEM.PUT(m.sb + tdEntry, SYSTEM.ADR(td.word[root]));
        m.tdescs[i] := SYSTEM.ADR(td.word[root]);
        j := 0;
        WHILE j < nofPointers DO File.ReadLInt(td.word[root+1+j]); INC(j) END;
        td.word[root+1+nofPointers] := -(nofPointers+1)*4;
      END
    END
  END ReadTypes;

  PROCEDURE InitType (i: LONGINT);
    VAR t, baseType, tag, entry: LONGINT; extlev, n, root, baseModNo: INTEGER; td: TDesc; baseMod: Module;
  BEGIN
    IF ~types[i].initialized THEN
      td := types[i].tdesc;
      baseModNo := types[i].baseMod;
      extlev := 0;
      root := types[i].root;
      IF baseModNo # -1 THEN
        IF baseModNo = 0 THEN
          t := 0;
	  WHILE (t < LEN(m^.tdescs^)) & (types[t].entry # types[i].baseEntry) DO INC(t) END;
	  InitType(t)
        END;

        IF baseModNo > 0 THEN
	  baseMod := m.imports[baseModNo-1];
	  FindTDescAdr(baseMod, types[i].baseEntry, types[i].baseEntry);
        ELSE
	  baseMod := m
        END;
        SYSTEM.GET(baseMod.sb + types[i].baseEntry, baseType);
        SYSTEM.GET(baseType + Tag0WordOffset*4, tag);
        WHILE tag # 0 DO
	  td.word[root + Tag0WordOffset - extlev] := tag;
	  INC(extlev);
	  SYSTEM.GET(baseType + (Tag0WordOffset - extlev)*4, tag);
        END;
        n := types[i].nofInhMeth;
        WHILE n > 0 DO DEC(n);
          entry := td.word[root + Mth0WordOffset - n];
          IF entry = 0 THEN
            SYSTEM.GET(baseType + (Mth0WordOffset - n)*4, td.word[root + Mth0WordOffset - n])
          END
        END
      END;
      td.word[3] := extlev;
      td.word[root + Tag0WordOffset - extlev] := SYSTEM.ADR(td.word[root]);
      types[i].initialized := TRUE
    END
  END InitType;

  PROCEDURE InitTypes;
  VAR
    i: LONGINT;
  BEGIN i := 0;
    WHILE i < LEN(m.tdescs^) DO InitType(i); INC(i) END
  END InitTypes;

  PROCEDURE BuildModuleBlock (m: Module; h: Header);
  VAR
    t, size, gvarSize: LONGINT;
    arrPtr: POINTER TO RECORD a, b, c, len, data: LONGINT END;
  BEGIN
    size := 16 * ((h.nofEntries*4 + 35) DIV 16
	    + (h.nofCmds*SIZE(Cmd) + 35) DIV 16 + (h.nofPtrs*4 + 35) DIV 16
	    + (h.nofTds*4 + 35) DIV 16 + ((h.nofImps)*4 + 35) DIV 16
	    + (h.dataSize + (-h.dataSize) MOD 8 + h.conSize + 35) DIV 16
	    + (h.codeSize + 35) DIV 16 + (h.refSize + 35) DIV 16);

    SYSTEM.NEW(arrPtr, size);
    SYSTEM.GET(SYSTEM.VAL(ADDRESS, arrPtr) - 4, t);

    SYSTEM.PUT(SYSTEM.ADR(m.entries), arrPtr);
    arrPtr.len := h.nofEntries;
    INC(SYSTEM.VAL(ADDRESS, arrPtr), LONG((h.nofEntries*4 + 35) DIV 16)*16);

    SYSTEM.PUT(SYSTEM.VAL(ADDRESS, arrPtr) - 4, t);
    SYSTEM.PUT(SYSTEM.ADR(m.cmds), arrPtr);
    arrPtr.len := h.nofCmds;
    INC(SYSTEM.VAL(ADDRESS, arrPtr), LONG((h.nofCmds*SIZE(Cmd) + 35) DIV 16)*16);

    SYSTEM.PUT(SYSTEM.VAL(ADDRESS, arrPtr) - 4, t);
    SYSTEM.PUT(SYSTEM.ADR(m.ptrTab), arrPtr);
    arrPtr.len := h.nofPtrs;
    INC(SYSTEM.VAL(ADDRESS, arrPtr), LONG((h.nofPtrs*4 + 35) DIV 16)*16);

    SYSTEM.PUT(SYSTEM.VAL(ADDRESS, arrPtr) - 4, t);
    SYSTEM.PUT(SYSTEM.ADR(m.tdescs), arrPtr);
    arrPtr.len := h.nofTds;
    INC(SYSTEM.VAL(ADDRESS, arrPtr), LONG((h.nofTds*4 + 35) DIV 16)*16);

    SYSTEM.PUT(SYSTEM.VAL(ADDRESS, arrPtr) - 4, t);
    SYSTEM.PUT(SYSTEM.ADR(m.imports), arrPtr);
    arrPtr.len := h.nofImps;
    INC(SYSTEM.VAL(ADDRESS, arrPtr), LONG(((h.nofImps)*4 + 35) DIV 16)*16);

    SYSTEM.PUT(SYSTEM.VAL(ADDRESS, arrPtr) - 4, t);
    SYSTEM.PUT(SYSTEM.ADR(m.data), arrPtr);
    gvarSize := h.dataSize + (-h.dataSize) MOD 8;
    m.sb := SYSTEM.ADR(arrPtr.data) + gvarSize;
    arrPtr.len := (gvarSize + h.conSize );
    INC(SYSTEM.VAL(ADDRESS, arrPtr), ((gvarSize + h.conSize + 35) DIV 16)*16);

    SYSTEM.PUT(SYSTEM.VAL(ADDRESS, arrPtr) - 4, t);
    SYSTEM.PUT(SYSTEM.ADR(m.code), arrPtr);
    arrPtr.len := h.codeSize;
    INC(SYSTEM.VAL(ADDRESS, arrPtr), ((h.codeSize + 35) DIV 16)*16);

    SYSTEM.PUT(SYSTEM.VAL(ADDRESS, arrPtr) - 4, t);
    SYSTEM.PUT(SYSTEM.ADR(m.refs), arrPtr);
    arrPtr.len := h.refSize
  END BuildModuleBlock;

BEGIN
  res := done;
  Clear;

  NEW(m);
  m.init := FALSE;
  m.refcnt := 0;
  m.term := NIL;

  File.ReadNum(symSize);
  File.SetPos(File.pos+symSize);

  File.ReadLInt(head.refSize);
  File.ReadInt(head.nofEntries);
  File.ReadInt(head.nofCmds);
  File.ReadInt(head.nofPtrs);
  File.ReadInt(head.nofTds);
  File.ReadInt(head.nofImps);
  File.ReadInt(head.nofDataLinks);
  File.ReadInt(head.nofLinks);
  File.ReadLInt(head.dataSize);
  ReadUnsigned(File,head.conSize);
  ReadUnsigned(File, head.codeSize);
  File.ReadString(m.name);
  importing := m.name;
  m.nofimp := -1;
  BuildModuleBlock(m, head);

  IF res # done THEN m := NIL; RETURN END;

  IF ~Expect(082X) THEN RETURN END;
  i := 0; t := SYSTEM.ADR(m.code[0]);
  WHILE i < head.nofEntries DO
    ReadUnsigned(File, e);
    m.entries[i] := t + e;
    INC(i)
  END;

  IF ~Expect(083X) THEN RETURN END;
  i := 0;
  WHILE i < head.nofCmds DO
    File.ReadString(m.cmds[i].name);
    ReadUnsigned(File, e);
    m.cmds[i].adr := t + e;
    INC(i)
  END;

  IF ~Expect(084X) THEN RETURN END;
  i := 0;
  t := m.sb;
  WHILE i < head.nofPtrs DO
    File.ReadLInt(k);
    m.ptrTab[i] := t + k;
    INC(i)
  END;

  IF ~Expect(085X) THEN RETURN END;
  IF head.nofImps # 0 THEN
    NEW(mods, head.nofImps);
    mno := 0;
    WHILE (mno < head.nofImps) & (res = done) DO
      File.ReadString(mods[mno]);
      Load(mods[mno], imp);
      IF res # done THEN RETURN END;
      m.imports[mno] := imp;
      INC(mno)
    END
  END;
  COPY(m.name, importing);

  IF ~Expect(08DX) THEN RETURN END;
  IF head.nofDataLinks # 0 THEN
    NEW(dataLinks, head.nofDataLinks);
    FOR i := 0 TO head.nofDataLinks-1 DO
      File.Read(dataLinks[i].mod);
      File.ReadInt(dataLinks[i].entry);
      File.ReadInt(dataLinks[i].nofFixups);
      IF dataLinks[i].nofFixups > 0 THEN
        NEW(dataLinks[i].offset, dataLinks[i].nofFixups);
        j := 0;
        WHILE j < dataLinks[i].nofFixups DO
          ReadUnsigned(File, dataLinks[i].offset[j]);
          INC(j)
        END
      ELSE
        dataLinks[i].offset := NIL
      END
    END
  END;

  IF ~Expect(086X) THEN RETURN END;
  IF head.nofLinks # 0 THEN
    NEW(linkTab, head.nofLinks);
    FOR i := 0 TO head.nofLinks-1 DO
      File.Read(linkTab[i].mod);
      File.Read(linkTab[i].entry);
      ReadUnsigned(File, linkTab[i].link)
    END
  END;

  IF ~Expect(087X) THEN RETURN END;
  i := 0; t := m.sb;
  WHILE i < head.conSize DO
    File.Read(ch);
    SYSTEM.PUT(t, ch);
    INC(t);
    INC(i)
  END;

  IF ~Expect(088X) THEN RETURN END;
  LoadExpBlock(File, m);

  IF ~Expect(089X) THEN RETURN END;
  File.ReadBytes(m.code^, head.codeSize);

  IF ~Expect(08AX) THEN RETURN END;
  CheckUseBlock(File, m);
  IF res # done THEN RETURN END;

  IF ~Expect(08BX) THEN RETURN END;
  ReadTypes(File);
  IF res # done THEN RETURN END;

  IF ~Expect(08CX) THEN RETURN END;
  File.ReadBytes(m.refs^, head.refSize);
  IF res = done THEN
    imp := FindMod(m.name);
    IF imp # NIL THEN m := imp; RETURN END;
    WHILE mno > 0 DO
      DEC(mno);
      imp := m.imports[mno];
      INC(imp.refcnt)
    END;
    SYSTEM.PUT (m.sb, SYSTEM.VAL (LONGINT, m));
    Fixup(m, linkTab);
    InitTypes;
    m.next := Kernel.modules;
    Kernel.modules := m;
    m.init := TRUE;
    SYSTEM.PUT(SYSTEM.ADR(body), SYSTEM.ADR(m.code[0]));
    body;
    res := done
  ELSE
    m := NIL
  END
END LoadModule;

PROCEDURE CheckName(File: Files.File; VAR name: ARRAY OF CHAR): BOOLEAN;
VAR
  fname: Files.FileName;
  i, j: LONGINT;
BEGIN
  fname := File.name;
  i := 0;
  j := 0;
  WHILE fname[i] # 0X DO
    IF fname[i] = Files.pathChar THEN j := i+1 END;
    INC(i)
  END;
  i := 0;
  WHILE (name[i] # 0X) & (fname[j] = name[i]) DO
    INC(i); INC(j)
  END;
  RETURN name[i] = 0X
END CheckName;

PROCEDURE Load (name: ARRAY OF CHAR; VAR m: Module);
VAR
  File: Files.File; i, j: INTEGER;
  fname: Files.FileName;
  tag: CHAR;
BEGIN
  m := FindMod(name);
  res := done;
  Clear;
  IF m = NIL THEN
    i := 0;
    WHILE name[i] # 0X DO fname[i] := name[i]; INC(i) END;
    j := 0;
    WHILE extension[j] # 0X DO fname[i] := extension[j]; INC(i); INC(j) END;
    fname[i] := 0X;

    NEW(File);

    IF ~ File.Open(fname) THEN
      COPY(fname, importing);
      res := fileNotFound;
      ErrMsg(res);
      RETURN
    ELSIF ~CheckName(File, fname) THEN
      COPY(fname, importing);
      res := invalidObjFile;
      ErrMsg(res);
      RETURN
    END;

    File.SetPos(0);
    File.Read(tag);

    IF tag = 0BBX THEN
      File.Read(tag);
      IF tag = 0ADX THEN File.Read(tag) END;
      IF (tag = ActiveTag) THEN
        LoadModule(File, m)
      ELSE
        COPY(name, importing);
        res := invalidObjFile;
        ErrMsg(res)
      END
    ELSE
      COPY(name, importing);
      res := invalidObjFile;
      ErrMsg(res)
    END;
    File.Close;
  ELSIF ~m.init THEN
    COPY(name, importing);
    res := cyclicImport;
    ErrMsg(res);
  END
END Load;

PROCEDURE ThisMod* (name: ARRAY OF CHAR): Module;
VAR
  excp: Kernel.ExcpFrm;
  mod: Module;
BEGIN
  excp.panic := FALSE;
  Kernel.LockAndTry(moduleCS, excp);
  Load(name, mod);
  Kernel.UnlockAndUnwind(moduleCS, excp);
  RETURN mod
END ThisMod;

PROCEDURE ThisCommand* (mod: Module; name: ARRAY OF CHAR): Command;
VAR
  i: LONGINT;
  cmd: Command;
BEGIN
  IF mod # NIL THEN
    Kernel.EnterCriticalSection(moduleCS);
    i := LEN(mod.cmds);
    WHILE i > 0 DO
      DEC(i);
      IF mod.cmds[i].name = name THEN
        res := done;
        Clear;
        cmd := SYSTEM.VAL(Command, mod.cmds[i].adr);
        Kernel.LeaveCriticalSection(moduleCS);
        RETURN cmd
      END
    END;
    COPY(mod.name, imported);
    COPY(name, importing);
    res := cmdNotFound;
    ErrMsg(res);
    Kernel.LeaveCriticalSection(moduleCS)
  END;
  RETURN NIL
END ThisCommand;

PROCEDURE Delete (m: Module);
VAR
  m1: Module;
BEGIN
  m1 := Kernel.modules;
  IF m = m1 THEN
    Kernel.modules := m.next
  ELSE
    WHILE m1.next # m DO m1 := m1.next END;
    m1.next := m.next
  END
END Delete;

PROCEDURE Free* (name: ARRAY OF CHAR; all: BOOLEAN);
VAR
  excp: Kernel.ExcpFrm;
  m, m1: Module;
  i, j: LONGINT;
BEGIN
  excp.panic := FALSE;
  Kernel.LockAndTry(moduleCS, excp);
  res := done;
  Clear;
  m := FindMod(name);
  IF m = NIL THEN
    COPY(name, importing);
    res := moduleNotFound;
    ErrMsg(res)
  ELSIF m.refcnt # 0 THEN
    importing := m.name;
    res := refCntNotZero;
    ErrMsg(res)
  ELSE
    Kernel.FinalizeModule(m);
    i := LEN(m.imports);
    WHILE i > 0 DO
      DEC(i);
      m1 := m.imports[i];
      DEC(m1.refcnt)
    END;
    IF all THEN
      i := LEN(m.imports);
      WHILE i > 0  DO
        DEC(i);
        m1 := m.imports[i];
        Free(m1.name, TRUE)
      END;
      res := done;
      Clear;
    END;
    Delete(m);
    i := SYSTEM.VAL(ADDRESS, m.entries);
    SYSTEM.GET(i-4, j);
    SYSTEM.GET(j, j);
    j := i + j-24;
    WHILE i < j DO
      SYSTEM.PUT32(i, 0CCCCCCCCH);
      INC(i, 4)
    END
  END;
  Kernel.UnlockAndUnwind(moduleCS, excp)
END Free;

PROCEDURE InstallTermHandler*(h: Kernel.Proc);
BEGIN
  Kernel.EnterCriticalSection(moduleCS);
  Kernel.InstallTermHandler(h);
  Kernel.LeaveCriticalSection(moduleCS)
END InstallTermHandler;

PROCEDURE *Shutdown();
BEGIN
  Kernel.DeleteCriticalSection(moduleCS)
END Shutdown;

BEGIN
  Kernel.InitializeCriticalSection(moduleCS);
  InstallTermHandler(Shutdown);
  extension := ".Obj";
END Modules.

